@@ -1,17 +1,16 @@
 import copy
 import re
-
+from io import StringIO, BytesIO
 
 from lxml import etree
 from reportlab.lib.colors import HexColor, black
 from reportlab.lib.units import mm
-from reportlab.platypus import TableStyle, CellStyle, PageBreak, Spacer
-from io import StringIO, BytesIO
+from reportlab.platypus import TableStyle, PageBreak, Spacer
 
 from .enhanced_paragraph.enhanced_paragraph import EnhancedParagraph
 from .enhanced_paragraph.style import EnhancedParagraphStyle
 from .enhanced_table.data import EnhancedTableData
-from .enhanced_table.enhanced_tables import OVERFLOW_ROW, EnhancedTable
+from .enhanced_table.enhanced_tables import OVERFLOW_ROW, EnhancedTable, HEADER_FOOTER
 from .png_images import insert_image, insert_obj
 from .svglib.svglib import SvgRenderer
 from .utils import DocTemplate, get_page_size_from_string, intcomma_currency, ColumnWidthPercentage
@@ -31,8 +30,6 @@ class ReportXML(object):
         (u'rdquo', u'”'),
         (u'quot', u'"'),
         (u'frasl', u'/'),
-        (u'lt', u'<'),
-        (u'gt', u'>'),
         (u'hellip', u'…'),
         (u'ndash', u'–'),
         (u'mdash', u'—'),
@@ -61,13 +58,15 @@ class ReportXML(object):
         (u'rsquo', u'’'),
     ]
 
-    def __init__(self, object_lookup=None, background_images=None, pager_kwargs=None):
+    def __init__(self, object_lookup=None, pager_kwargs=None):
         self.styles = {}
         if object_lookup is not None:
             self.object_lookup = object_lookup
         else:
             self.object_lookup = {}
-        self.background_images = background_images
+        self.background_image_first = None
+        self.background_image_remaining = None
+        self.background_image_footer = None
 
         self.styles = {}
         self.border_left_first = 0
@@ -85,7 +84,13 @@ class ReportXML(object):
         else:
             self.pager_kwargs = pager_kwargs
 
-    def load_xml_and_make_pdf(self, xml, add_doctype=True):
+    def load_xml_and_make_pdf(self, xml, add_doctype=True, background_image_first=None,
+                              background_image_remaining=None, background_image_footer=None):
+
+        self.background_image_first = background_image_first
+        self.background_image_remaining = background_image_remaining
+        self.background_image_footer = background_image_footer
+
         parser = etree.XMLParser(remove_blank_text=True, resolve_entities=False)
 
         if add_doctype:
@@ -168,6 +173,9 @@ class ReportXML(object):
                               border_right_continuation=self.border_right_continuation * mm,
                               border_top_continuation=self.border_top_continuation * mm,
                               border_bottom_continuation=self.border_bottom_continuation * mm,
+                              background_image_first=self.background_image_first,
+                              background_image_remaining=self.background_image_remaining,
+                              background_image_footer=self.background_image_footer,
                               **self.pager_kwargs)
 
     def canvasmaker(self, *args, **kwargs):
@@ -181,6 +189,9 @@ class ReportXML(object):
                               border_right_continuation=self.border_right_continuation * mm,
                               border_top_continuation=self.border_top_continuation * mm,
                               border_bottom_continuation=self.border_bottom_continuation * mm,
+                              background_image_first=self.background_image_first,
+                              background_image_remaining=self.background_image_remaining,
+                              background_image_footer=self.background_image_footer,
                               **self.pager_kwargs,
                               **kwargs)
 
@@ -233,13 +244,14 @@ class ReportXML(object):
         main_styles = []
         main_span = {}
 
-        header_data = []
-        header_styles = []
-        header_row_height = []
+        headers = []
+        footers = []
+
+        headers_index = []
+        footers_index = []
+        current_header_index = None
+        current_footer_index = None
 
-        footer_data = []
-        footer_styles = []
-        footer_row_height = []
         row_heights = []
         keep_data = []
         other_styles = {}
@@ -285,14 +297,19 @@ class ReportXML(object):
                 if held_row_span > 1 or min_rows_top > 0:
                     for _ in range(overflow_row_count + 1):
                         keep_data.append(1)
+                        headers_index.append(current_header_index)
+                        footers_index.append(current_footer_index)
                     min_rows_top -= 1
                 else:
                     for _ in range(overflow_row_count + 1):
                         keep_data.append(0)
+                        headers_index.append(current_header_index)
+                        footers_index.append(current_footer_index)
                 held_row_span -= 1
                 row_count += overflow_row_count
             elif element.tag == 'keep':
                 local_keep_data = []
+
                 for tr in element:
                     row_count += 1
                     _, overflow_row_count = self.process_tr(tr_element=tr,
@@ -309,17 +326,71 @@ class ReportXML(object):
                     if min_rows_top > 0:
                         min_rows_top -= 1
 
-                    for _ in range(overflow_row_count + 1):
-                        local_keep_data.append(2)
+                    for index, _ in enumerate(range(overflow_row_count + 1)):
+                        headers_index.append(current_header_index)
+                        footers_index.append(current_footer_index)
+                        if index == 0:
+                            local_keep_data.append(2)
+                        else:
+                            local_keep_data.append(3)
+                if len(local_keep_data) > 0:
+                    local_keep_data[-1] = 3
+                keep_data += local_keep_data
 
-                    keep_data += local_keep_data
+            elif element.tag == 'no_headers':
+                current_header_index = None
 
-            elif element.tag == 'header':
-                for tr in element:
-                    self.process_header_or_footer_tr(tr, header_data, header_styles, header_row_height)
-            elif element.tag == 'footer':
-                for tr in element:
-                    self.process_header_or_footer_tr(tr, footer_data, footer_styles, footer_row_height)
+            elif element.tag == 'no_footers':
+                current_footer_index = None
+
+            elif element.tag in ['header', 'footer']:
+                is_header = element.tag == 'header'
+
+                if is_header and int(element.get('output', "0")) == 1:
+                    for tr in element:
+                        row_count += 1
+                        _, overflow_row_count = self.process_tr(tr_element=tr,
+                                                                data=main_data,
+                                                                styles=main_styles,
+                                                                other_table_styles=other_styles,
+                                                                row_heights=row_heights,
+                                                                row_count=row_count,
+                                                                span=main_span,
+                                                                rows_variables=rows_variables,
+                                                                variables=variables,
+                                                                col_widths=col_widths,
+                                                                table_width=table_width)
+                header_footer_span = {}
+                header_footer_data = []
+                header_footer_commands = []
+                header_footer_row_height = []
+                temp_rows_variables = []
+
+                for row_index, tr in enumerate(element, HEADER_FOOTER):
+                    self.process_tr(tr_element=tr,
+                                    data=header_footer_data,
+                                    styles=header_footer_commands,
+                                    other_table_styles=other_styles,
+                                    row_heights=header_footer_row_height,
+                                    row_count=row_index,
+                                    span=header_footer_span,
+                                    rows_variables=temp_rows_variables,
+                                    variables=variables,
+                                    col_widths=col_widths,
+                                    table_width=table_width,
+                                    default_row_height=35,
+                                    is_header_or_footer=True)
+
+                enhanced_table_data = EnhancedTableData(row_data=header_footer_data,
+                                                        row_heights=header_footer_row_height,
+                                                        commands=header_footer_commands)
+
+                if is_header:
+                    current_header_index = len(headers)
+                    headers.append(enhanced_table_data)
+                else:
+                    current_footer_index = len(footers)
+                    footers.append(enhanced_table_data)
 
         length = len(keep_data)
         for x in range(1, min_rows_bottom + 1):
@@ -327,24 +398,20 @@ class ReportXML(object):
                 break
             keep_data[length - x] = 1
 
-        header = EnhancedTableData(row_data=header_data,
-                                   row_heights=header_row_height,
-                                   cell_styles=header_styles)
-
-        footer = EnhancedTableData(row_data=footer_data,
-                                   row_heights=footer_row_height,
-                                   cell_styles=footer_styles)
-
         h_align, v_align = self.get_alignment_details(main_styles)
 
         new_column_widths = self.process_column_widths(col_widths, table_width)
 
         if main_data:
-            t = EnhancedTable({'row_data': main_data, 'row_variables': rows_variables, 'keep_with_next': keep_data},
+            t = EnhancedTable(table_data={'row_data': main_data,
+                                          'row_variables': rows_variables,
+                                          'keep_with_next': keep_data,
+                                          'headers_index': headers_index,
+                                          'footers_index': footers_index},
                               row_heights=row_heights,
                               repeat_rows=0,
-                              header=header,
-                              footer=footer,
+                              headers=headers,
+                              footers=footers,
                               h_align=h_align,
                               v_align=v_align,
                               col_widths=new_column_widths,
@@ -390,43 +457,9 @@ class ReportXML(object):
         x, y = x * mm, 222 - y * mm
         return x, y
 
-    def process_header_or_footer_tr(self, tr, data, styles, row_heights):
-        row_data = []
-        row_styles = []
-
-        row_css = self.get_css_from_style_attribute(tr)
-
-        row_heights.append(int(tr.get('row_height', 35)))
-        other_styles = {}
-
-        for td_element in tr.iter('td'):
-            td_css = row_css + self.get_css_from_style_attribute(td_element)
-
-            if len(td_element) > 0 and td_element[0].tag[-8:] == 'currency':
-                variable_name = td_element[0].get('variable')
-                if variable_name is not None:
-                    value = '%%(%s__currency)s' % variable_name
-                    row_data.append(value)
-                else:
-                    value = int(td_element[0].get('value'))
-                    number_string = "%.2f" % (float(value) / 100.0)
-                    row_data.append('%s' % (intcomma_currency(number_string)))
-            elif len(td_element) > 0:
-                new_styles = copy.deepcopy(styles)
-                self.convert_css_to_style(td_css, new_styles, other_styles)
-                style = self.process_css_for_table_paragraph_style(new_styles, None, None)
-                xml = etree.tostring(td_element)
-                row_data.append(EnhancedParagraph(xml, style, css_classes=self.styles))
-            elif td_element.text is not None:
-                row_data.append(td_element.text)
-            else:
-                row_data.append('')
-            row_styles.append(self.process_css_for_header_or_footer(td_css))
-        styles.append(row_styles)
-        data.append(row_data)
-
     def process_tr(self, tr_element, data, styles, other_table_styles,
-                   row_heights, row_count, span, rows_variables, variables, col_widths, table_width):
+                   row_heights, row_count, span, rows_variables, variables, col_widths, table_width,
+                   default_row_height=None, is_header_or_footer=False):
         row_data = []
         other_styles = {}
 
@@ -444,6 +477,12 @@ class ReportXML(object):
                 variables[variable] = value
                 row_variables[variable] = value
 
+        for variable_element in tr_element.iter('variable'):
+            variable = variable_element.attrib.get('name', '')
+            value = variable_element.text
+            variables[variable] = value
+            row_variables[variable] = value
+
         for variables_element in tr_element.iter('currency_variables'):
             symbol = variables_element.attrib.get('symbol', '')
 
@@ -566,35 +605,45 @@ class ReportXML(object):
                 display_object = EnhancedParagraph(display_str, style, css_classes=self.styles)
 
             elif len(td_element) > 0 and td_element[0].tag[-8:] == 'currency':
-
                 variable_name = td_element[0].get('variable')
+                if is_header_or_footer:
+                    variable_name = td_element[0].get('variable')
+                    if variable_name is not None:
+                        display_object = '%%(%s__currency)s' % variable_name
+                    else:
+                        value = int(td_element[0].get('value'))
+                        number_string = "%.2f" % (float(value) / 100.0)
+                        display_object = '%s' % (intcomma_currency(number_string))
 
-                if variable_name is not None:
-                    symbol = ''
-                    value = variables[variable_name]
-                    if symbol == '':
-                        symbol = variables.get(variable_name + '__symbol', '')
                 else:
-                    symbol = td_element[0].get('symbol', '')
-                    value = td_element[0].get('value')
+                    if variable_name is not None:
+                        symbol = ''
+                        value = variables.get(variable_name)
+                        if symbol == '':
+                            symbol = variables.get(variable_name + '__symbol', '')
+                    else:
+                        symbol = td_element[0].get('symbol', '')
+                        value = td_element[0].get('value')
 
-                add_to_variable_name = td_element[0].get('add_to')
+                    add_to_variable_name = td_element[0].get('add_to')
 
-                if add_to_variable_name is not None:
-                    if symbol == '':
-                        symbol = variables.get(add_to_variable_name + '__symbol', '')
+                    if add_to_variable_name is not None:
+                        if symbol == '':
+                            symbol = variables.get(add_to_variable_name + '__symbol', '')
 
-                    variables[add_to_variable_name] += float(value)
-                    row_variables[add_to_variable_name] = variables.get(add_to_variable_name)
-                    currency_variable = add_to_variable_name + '__currency'
-                    if currency_variable in variables:
-                        currency_value = symbol + intcomma_currency(float(variables.get(add_to_variable_name)) / 100.0)
+                        variables[add_to_variable_name] += float(value)
+                        row_variables[add_to_variable_name] = variables.get(add_to_variable_name)
+                        currency_variable = add_to_variable_name + '__currency'
+                        if currency_variable in variables:
+                            currency_value = symbol + intcomma_currency(float(variables.get(add_to_variable_name)) / 100.0)
 
-                        row_variables[currency_variable] = currency_value
-                        variables[currency_variable] = currency_value
-
-                number_string = "%.2f" % (float(value) / 100.0)
-                display_object = '%s%s' % (symbol, intcomma_currency(number_string))
+                            row_variables[currency_variable] = currency_value
+                            variables[currency_variable] = currency_value
+                    if value is not None:
+                        number_string = "%.2f" % (float(value) / 100.0)
+                        display_object = '%s%s' % (symbol, intcomma_currency(number_string))
+                    else:
+                        display_object = ''
 
             elif len(td_element) > 0:
                 xml = etree.tostring(td_element, pretty_print=True)
@@ -649,7 +698,13 @@ class ReportXML(object):
 
         data.append(row_data)
 
-        row_heights.append(self.get_row_height(other_table_styles, other_styles))
+        row_height = self.get_row_height(other_table_styles, other_styles)
+        if row_height is None:
+            row_height = tr_element.get('row_height', default_row_height)
+            if row_height is not None:
+                row_height = float(row_height)
+
+        row_heights.append(row_height)
 
         for overflow_row in overflow_rows:
             data.append(overflow_row)
@@ -750,7 +805,6 @@ class ReportXML(object):
     @staticmethod
     def get_padding_for_cell(styles, start_col=0, start_row=0, end_col=-1, end_row=-1):
         start_tuple = (start_col, start_row)
-        end_tuple = (end_col, end_row)
         padding = 0
         found_padding = False
         for style in styles:
@@ -780,49 +834,6 @@ class ReportXML(object):
             css += style_tag
         return css
 
-    @staticmethod
-    def process_css_for_header_or_footer(css):
-        cell_style = CellStyle('header_footer')
-        css = css.replace('\r', '').replace('\n', '')
-
-        styles_list = css.split(';')
-        for style in styles_list:
-            if style == '':
-                continue
-            style_type, style_detail = style.split(':')
-            style_type = style_type.lower().lstrip().rstrip()
-            if style_type in ('font_name',
-                              'face',
-                              'font'):
-                cell_style.fontname = style_detail
-            elif style_type in ('size',
-                                'font_size'):
-                cell_style.fontsize = float(style_detail)
-            elif style_type == 'leading':
-                cell_style.leading = style_detail
-            elif style_type == 'text_color':
-                cell_style.color = HexColor(style_detail)
-            elif style_type in ('align',
-                                'alignment'):
-                cell_style.alignment = style_detail.upper()
-            elif style_type == 'valign':
-                cell_style.valign = style_detail.upper()
-
-            elif style_type == 'halign':
-                cell_style.halign = style_detail
-            elif style_type == 'left_padding':
-                cell_style.leftPadding = int(style_detail)
-            elif style_type == 'right_padding':
-                cell_style.rightPadding = int(style_detail)
-            elif style_type == 'top_padding':
-                cell_style.topPadding = int(style_detail)
-            elif style_type == 'bottom_padding':
-                cell_style.bottomPadding = int(style_detail)
-            elif style_type == 'background':
-                cell_style.background = HexColor(style_detail)
-
-        return cell_style
-
     @staticmethod
     def process_css_for_table_paragraph_style(css, row_count, col_count):
         paragraph_style = EnhancedParagraphStyle('paragraph_style')
@@ -938,10 +949,7 @@ class ReportXML(object):
                         elif tag_name == working_tags[-1][0]:
                             working_tags.pop()
                             if tag_name == 'td' and len(working_tags) == 0:
-                                if len(overflow_rows) > 0:
-                                    return overflow_rows[0], overflow_rows[1:]
-                                else:
-                                    return '', []
+                                break
                     else:
                         working_tags.append((tag_name, tag))
 
